#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.14 11:00:00                  #
# ================================================== #
import os.path
import subprocess

from ..base_plugin import BasePlugin


class Plugin(BasePlugin):
    def __init__(self):
        super(Plugin, self).__init__()
        self.id = "cmd_code_interpreter"
        self.name = "Command: Code Interpreter"
        self.description = "Provides Python code execution"
        self.window = None
        self.order = 100
        self.allowed_cmds = [
            "code_execute",
            "code_execute_file",
            "sys_exec"
        ]
        self.init_options()

    def init_options(self):
        """
        Initializes options
        """
        # cmd enable/disable
        self.add_option("python_cmd_tpl", "text", 'python3 {filename}',
                        "Python command template",
                        "Python command template to execute, use {filename} for filename placeholder")
        self.add_option("cmd_code_execute", "bool", True,
                        "Enable: Python Code Generate and Execute",
                        "Allows Python code execution (generate and execute from file)")
        self.add_option("cmd_code_execute_file", "bool", True,
                        "Enable: Python Code Execute (File)",
                        "Allows Python code execution from existing file")
        self.add_option("cmd_sys_exec", "bool", True,
                        "Enable: System Command Execute",
                        "Allows system commands execution")

        # cmd syntax (prompt/instruction)
        self.add_option("syntax_code_execute", "textarea", '"code_execute": create and execute Python code, params: '
                                                           '"filename", "code"',
                        "Syntax: code_execute",
                        "Syntax for Python code execution (generate and execute from file)", advanced=True)
        self.add_option("syntax_code_execute_file", "textarea", '"code_execute_file": execute Python code from '
                                                                'existing file, params: "filename"',
                        "Syntax: code_execute_file",
                        "Syntax for Python code execution from existing file", advanced=True)
        self.add_option("syntax_sys_exec", "textarea", '"sys_exec": execute ANY system command, script or application '
                                                       'in user\'s environment, params: "command"',
                        "Syntax: sys_exec",
                        "Syntax for system commands execution", advanced=True)

    def setup(self):
        """
        Returns available config options

        :return: config options
        """
        return self.options

    def attach(self, window):
        """
        Attaches window

        :param window: Window
        """
        self.window = window

    def on_user_send(self, text):
        """
        Event: On user send text

        :param text: Text
        :return: Text
        """
        return text

    def on_ctx_begin(self, ctx):
        """
        Event: On new context begin

        :param ctx: Context
        :return: Context
        """
        return ctx

    def on_ctx_end(self, ctx):
        """
        Event: On context end

        :param ctx: Context
        :return: Context
        """
        return ctx

    def on_system_prompt(self, prompt):
        """
        Event: On prepare system prompt

        :param prompt: Prompt
        :return: Prompt
        """
        return prompt

    def on_ai_name(self, name):
        """
        Event: On set AI name

        :param name: Name
        :return: Name
        """
        return name

    def on_user_name(self, name):
        """
        Event: On set username

        :param name: Name
        :return: Name
        """
        return name

    def on_enable(self):
        """Event: On plugin enable"""
        pass

    def on_disable(self):
        """Event: On plugin disable"""
        pass

    def on_input_before(self, text):
        """
        Event: Before input

        :param text: Text
        :return: Text
        """
        return text

    def on_ctx_before(self, ctx):
        """
        Event: Before ctx

        :param ctx: Text

        """
        return ctx

    def on_ctx_after(self, ctx):
        """
        Event: After ctx

        :param ctx: ctx
        :return: ctx
        """
        return ctx

    def is_cmd_allowed(self, cmd):
        """
        Checks if cmd is allowed

        :param cmd: Command
        :return: True if allowed
        """
        key = "cmd_" + cmd
        if key in self.options and self.options[key]["value"] is True:
            return True
        return False

    def log(self, msg):
        """
        Logs message to console

        :param msg: Message to log
        """
        self.window.log('[CMD] ' + str(msg))
        print('[CMD] ' + str(msg))

    def cmd_syntax(self, syntax):
        """
        Event: On cmd syntax prepare

        :param syntax: Syntax
        :return: Syntax
        """
        for item in self.allowed_cmds:
            if self.is_cmd_allowed(item):
                key = "syntax_" + item
                if key in self.options:
                    syntax += "\n" + self.options[key]["value"]
        return syntax

    def cmd(self, ctx, cmds):
        """
        Execute command
        :param ctx: Context
        :param cmds: Commands list
        :return: Context
        """
        msg = None
        for item in cmds:
            try:
                if item["cmd"] in self.allowed_cmds:
                    # prepare request item for result
                    request_item = {"cmd": item["cmd"]}
                    ctx.reply = True  # send result message
                    result = None

                    # code_execute (from existing file)
                    if item["cmd"] == "code_execute_file" and self.is_cmd_allowed("code_execute_file"):
                        try:
                            # execute code from file
                            msg = "Executing Python file: {}".format(item["params"]['filename'])
                            self.log(msg)
                            path = os.path.join(self.window.config.path, 'output', item["params"]['filename'])

                            # check if file exists
                            if not os.path.isfile(path):
                                msg = "File not found: {}".format(item["params"]['filename'])
                                ctx.results.append({"request": request_item, "result": "File not found"})
                                continue

                            # run code
                            cmd = self.options['python_cmd_tpl']['value'].format(filename=path)
                            self.log("Running command: {}".format(cmd))
                            process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                                                       stderr=subprocess.PIPE)
                            stdout, stderr = process.communicate()
                            if stdout:
                                result = stdout.decode("utf-8")
                                self.log("STDOUT: {}".format(result))
                            if stderr:
                                result = stderr.decode("utf-8")
                                self.log("STDERR: {}".format(result))
                            if result is None:
                                result = "No result (STDOUT/STDERR empty)"
                                self.log(result)
                            ctx.results.append({"request": request_item, "result": result})
                        except Exception as e:
                            ctx.results.append({"request": request_item, "result": "Error: {}".format(e)})
                            self.log("Error: {}".format(e))

                    # code_execute (generate and execute)
                    elif item["cmd"] == "code_execute" and self.is_cmd_allowed("code_execute"):
                        try:
                            # write code to file
                            msg = "Saving Python file: {}".format(item["params"]['filename'])
                            self.log(msg)
                            path = os.path.join(self.window.config.path, 'output', item["params"]['filename'])
                            data = item["params"]['code']
                            with open(path, 'w', encoding="utf-8") as file:
                                file.write(data)
                                file.close()
                            # run code
                            cmd = self.options['python_cmd_tpl']['value'].format(filename=path)
                            self.log("Running command: {}".format(cmd))
                            process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                                                       stderr=subprocess.PIPE)
                            stdout, stderr = process.communicate()
                            if stdout:
                                result = stdout.decode("utf-8")
                                self.log("STDOUT: {}".format(result))
                            if stderr:
                                result = stderr.decode("utf-8")
                                self.log("STDERR: {}".format(result))
                            if result is None:
                                result = "No result (STDOUT/STDERR empty)"
                                self.log(result)
                            ctx.results.append({"request": request_item, "result": result})
                        except Exception as e:
                            ctx.results.append({"request": item, "result": "Error: {}".format(e)})
                            self.log("Error: {}".format(e))

                    # sys_exec
                    elif item["cmd"] == "sys_exec" and self.is_cmd_allowed("sys_exec"):
                        try:
                            # execute system command
                            msg = "Executing system command: {}".format(item["params"]['command'])
                            self.log(msg)
                            self.log("Running command: {}".format(item["params"]['command']))
                            process = subprocess.Popen(item["params"]['command'], shell=True, stdout=subprocess.PIPE,
                                                       stderr=subprocess.PIPE)
                            stdout, stderr = process.communicate()
                            if stdout:
                                result = stdout.decode("utf-8")
                                self.log("STDOUT: {}".format(result))
                            if stderr:
                                result = stderr.decode("utf-8")
                                self.log("STDERR: {}".format(result))
                            if result is None:
                                result = "No result (STDOUT/STDERR empty)"
                                self.log(result)
                            ctx.results.append({"request": request_item, "result": result})
                        except Exception as e:
                            ctx.results.append({"request": request_item, "result": "Error: {}".format(e)})
                            self.log("Error: {}".format(e))
            except Exception as e:
                ctx.results.append({"request": item, "result": "Error: {}".format(e)})
                ctx.reply = True  # send result message
                self.log("Error: {}".format(e))

        if msg is not None:
            self.window.statusChanged.emit(msg)
        return ctx



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.13 19:00:00                  #
# ================================================== #

from ..base_plugin import BasePlugin


class Plugin(BasePlugin):
    def __init__(self):
        super(Plugin, self).__init__()
        self.id = "self_loop"
        self.name = "Self Loop"
        self.description = "Allows to execute self talk (AI to AI) loop and connect output to input."
        self.iteration = 0
        self.prev_output = None
        self.window = None
        self.order = 9998
        self.init_options()

    def init_options(self):
        """
        Initializes options
        """
        self.add_option("iterations", "int", 3,
                        "Iterations",
                        "How many iterations to run? 0 = infinite.\nWARNING: setting this to 0 can cause a lot of "
                        "requests and heavy tokens usage!",
                        min=0, max=100, multiplier=1, step=1, slider=True)
        self.add_option("clear_output", "bool", True,
                        "Clear context output",
                        "If enabled, previous context output will be cleared before sending new input.")
        self.add_option("reverse_roles", "bool", True,
                        "Reverse roles between iterations",
                        "If enabled, roles will be reversed between iterations.")

    def setup(self):
        """
        Returns available config options

        :return: config options
        """
        return self.options

    def attach(self, window):
        """
        Attaches window

        :param window: Window
        """
        self.window = window

    def on_user_send(self, text):
        """
        Event: On user send text

        :param text: Text
        :return: Text
        """
        self.iteration = 0
        self.prev_output = None
        return text

    def on_ctx_begin(self, ctx):
        """
        Event: On new context begin

        :param ctx: Context
        :return: Context
        """
        return ctx

    def on_ctx_end(self, ctx):
        """
        Event: On context end

        :param ctx: Context
        :return: Context
        """
        iterations = int(self.options["iterations"]["value"])
        if iterations == 0 or self.iteration < iterations:
            self.iteration += 1
            if self.prev_output is not None and self.prev_output != "":
                self.window.log(
                    "Plugin: self_loop:on_ctx_end [sending prev_output...]: {}".format(self.prev_output))  # log
                self.window.controller.input.send(self.prev_output)
        return ctx

    def on_system_prompt(self, prompt):
        """
        Event: On prepare system prompt

        :param prompt: Prompt
        :return: Prompt
        """
        return prompt

    def on_ai_name(self, name):
        """
        Event: On set AI name

        :param name: Name
        :return: Name
        """
        return name

    def on_user_name(self, name):
        """
        Event: On set user name

        :param name: Name
        :return: Name
        """
        return name

    def on_enable(self):
        """Event: On plugin enable"""
        pass

    def on_disable(self):
        """Event: On plugin disable"""
        pass

    def on_input_before(self, text):
        """
        Event: Before input

        :param text: Text
        """
        return text

    def on_ctx_before(self, ctx):
        """
        Event: Before ctx

        :param ctx: Text
        """
        if self.iteration > 0 and self.iteration % 2 != 0 and self.options["reverse_roles"]["value"]:
            self.window.log("Plugin: self_loop:on_ctx_before [before]: {}".format(ctx.dump()))  # log
            tmp_input_name = ctx.input_name
            tmp_output_name = ctx.output_name
            ctx.input_name = tmp_output_name
            ctx.output_name = tmp_input_name
            self.window.log("Plugin: self_loop:on_ctx_before [after]: {}".format(ctx.dump()))  # log
        return ctx

    def on_ctx_after(self, ctx):
        """
        Event: After ctx

        :param ctx: ctx
        """
        self.prev_output = ctx.output
        if self.options["clear_output"]["value"]:
            self.window.log("Plugin: self_loop:on_ctx_after [before]: {}".format(ctx.dump()))  # log
            ctx.output = ""
            self.window.log("Plugin: self_loop:on_ctx_after [after]: {}".format(ctx.dump()))  # log
        return ctx



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.11 23:00:00                  #
# ================================================== #
import os
import threading
import time
import webbrowser

from PySide6.QtCore import QObject, Signal, Slot

from ..utils import trans
from ..assistants import Assistants


class Assistant:
    def __init__(self, window=None):
        """
        Presets controller

        :param window: main window object
        """
        self.window = window
        self.assistants = Assistants(self.window.config)
        self.thread_run = None
        self.thread_run_started = False
        self.force_stop = False

    def setup(self):
        """Setup assistants"""
        self.assistants.load()
        self.update()

    def update(self):
        """Updates assistants list"""
        self.update_list()
        self.update_uploaded()
        self.select_assistant_by_current()

    def update_list(self):
        """Updates assistants list"""
        items = self.assistants.get_all()
        self.window.ui.toolbox.update_list_assistants('assistants', items)

    def update_uploaded(self):
        """Updates uploaded files list"""
        assistant_id = self.window.config.data['assistant']
        if assistant_id is None or assistant_id == "":
            return
        assistant = self.assistants.get_by_id(assistant_id)
        items = assistant.files
        self.window.ui.attachments_uploaded.update_list('attachments_uploaded', items)
        self.update_tab_label()

    def update_tab_label(self):
        """
        Updates tab label

        :param mode: mode
        """
        assistant_id = self.window.config.data['assistant']
        if assistant_id is None or assistant_id == "":
            self.window.tabs['input'].setTabText(2, trans('attachments_uploaded.tab'))
            return

        assistant = self.assistants.get_by_id(assistant_id)
        items = assistant.files
        num_files = len(items)
        suffix = ''
        if num_files > 0:
            suffix = f' ({num_files})'
        self.window.tabs['input'].setTabText(2, trans('attachments_uploaded.tab') + suffix)

    def select(self, idx):
        """
        Selects assistant on the list

        :param idx: IDx on the list
        """
        # mark assistant as selected
        id = self.assistants.get_by_idx(idx)
        self.window.config.data['assistant'] = id

        # update attachments list with list of attachments from assistant
        mode = self.window.config.data['mode']
        assistant = self.assistants.get_by_id(id)
        self.window.controller.attachment.import_from_assistant(mode, assistant)
        self.window.controller.attachment.update()

    def update_field(self, id, value, assistant_id=None, current=False):
        """
        Updates assistant field from editor

        :param id: field id
        :param value: field value
        :param assistant_id: assistant ID
        :param current: if True, updates current assistant
        """
        if assistant_id is not None and assistant_id != "":
            if self.assistants.has(assistant_id):
                assistant = self.assistants.get_by_id(assistant_id)
                if id == 'assistant.name':
                    assistant.name = value
                elif id == 'assistant.description':
                    assistant.description = value
                elif id == 'assistant.instructions':
                    assistant.instructions = value
                elif id == 'assistant.model':
                    assistant.model = value

        self.window.controller.ui.update()

    def edit(self, idx=None):
        """
        Opens assistant editor

        :param idx: assistant index (row index)
        """
        id = None
        if idx is not None:
            id = self.assistants.get_by_idx(idx)

        self.init_editor(id)
        self.window.ui.dialogs.open_editor('editor.assistants', idx)

    def init_editor(self, id=None):
        """
        Initializes assistant editor

        :param id: assistant ID
        """
        assistant = self.assistants.create()
        assistant.model = "gpt-4-1106-preview"  # default model

        # if editing existing assistant
        if id is not None and id != "":
            if self.assistants.has(id):
                assistant = self.assistants.get_by_id(id)
        else:
            # default instructions
            assistant.instructions = 'You are a helpful assistant.'

        if assistant.name is None:
            assistant.name = ""
        if assistant.description is None:
            assistant.description = ""
        if assistant.instructions is None:
            assistant.instructions = ""
        if assistant.model is None:
            assistant.model = ""

        self.window.config_option['assistant.id'].setText(id)
        self.config_change('assistant.name', assistant.name, 'assistant.editor')
        self.config_change('assistant.description', assistant.description, 'assistant.editor')
        self.config_change('assistant.instructions', assistant.instructions, 'assistant.editor')
        self.config_change('assistant.model', assistant.model, 'assistant.editor')

        if assistant.has_tool('code_interpreter'):
            self.config_toggle('assistant.tool.code_interpreter', True, 'assistant.editor')
        else:
            self.config_toggle('assistant.tool.code_interpreter', False, 'assistant.editor')

        if assistant.has_tool('retrieval'):
            self.config_toggle('assistant.tool.retrieval', True, 'assistant.editor')
        else:
            self.config_toggle('assistant.tool.retrieval', False, 'assistant.editor')

        # restore functions
        if assistant.has_functions():
            functions = assistant.get_functions()
            values = []
            for function in functions:
                values.append({"name": function['name'], "params": function['params'], "desc": function['desc']})
            self.window.config_option['assistant.tool.function'].items = values
            self.window.config_option['assistant.tool.function'].model.updateData(values)
        else:
            self.window.config_option['assistant.tool.function'].items = []
            self.window.config_option['assistant.tool.function'].model.updateData([])

        # set focus to name field
        self.window.config_option['assistant.name'].setFocus()

    def save(self):
        """
        Saves assistant
        """
        created = False
        id = self.window.config_option['assistant.id'].text()
        if id is None or id == "" or not self.assistants.has(id):
            assistant = self.create()  # id created in API
            if assistant is None:
                print("Assistant not created")
                return
            id = assistant.id
            self.assistants.add(assistant)
            self.window.config_option['assistant.id'].setText(id)
            created = True
        else:
            assistant = self.assistants.get_by_id(id)

        # assign data from fields to assistant
        self.assign_data(assistant)

        # update in API
        if not created:
            self.assistant_update(assistant)

        # save file
        self.assistants.save()
        self.window.controller.model.update()
        self.update()

        self.window.ui.dialogs.close('editor.assistants')
        self.window.set_status(trans('status.assistant.saved'))

    def create(self):
        """
        Creates assistant
        """
        assistant = self.assistants.create()
        self.assign_data(assistant)
        try:
            return self.window.gpt.assistant_create(assistant)
        except Exception as e:
            self.window.ui.dialogs.alert(str(e))

    def assistant_update(self, assistant):
        """
        Update assistant
        """
        self.assign_data(assistant)
        try:
            return self.window.gpt.assistant_update(assistant)
        except Exception as e:
            self.window.ui.dialogs.alert(str(e))

    def import_assistants(self, force=False):
        """
        Imports all remote assistants from API
        """
        if not force:
            self.window.ui.dialogs.confirm('assistant_import', '', trans('confirm.assistant.import'))
            return

        try:
            # import assistants
            items = self.assistants.get_all()
            self.window.gpt.assistant_import(items)
            self.assistants.items = items
            self.assistants.save()

            # import uploaded files
            for assistant_id in self.assistants.items:
                assistant = self.assistants.get_by_id(assistant_id)
                self.import_files(assistant)
            # status
            self.window.set_status("Imported assistants: " + str(len(items)))
        except Exception as e:
            print("Error importing assistants")
            print(e)
            self.window.ui.dialogs.alert(str(e))
        self.update()

    def import_files(self, assistant):
        """
        Imports assistant files

        :param assistant: assistant
        """
        try:
            files = self.window.gpt.assistant_file_list(assistant.id)
            self.assistants.import_files(assistant, files)
            self.assistants.save()
            self.update()
            self.window.set_status("Imported files: " + str(len(files)))
        except Exception as e:
            print("Error importing assistant files")
            print(e)
            self.window.ui.dialogs.alert(str(e))

    def assign_data(self, assistant):
        """
        Assigns data from fields to assistant

        :param assistant: assistant
        """
        assistant.name = self.window.config_option['assistant.name'].text()
        assistant.model = self.window.config_option['assistant.model'].text()
        assistant.description = self.window.config_option['assistant.description'].text()
        assistant.instructions = self.window.config_option['assistant.instructions'].toPlainText()
        assistant.tools = {
            'code_interpreter': self.window.config_option['assistant.tool.code_interpreter'].box.isChecked(),
            'retrieval': self.window.config_option['assistant.tool.retrieval'].box.isChecked(),
            'function': [],  # functions are assigned separately
        }

        # assign functions
        functions = []
        for function in self.window.config_option['assistant.tool.function'].items:
            name = function['name']
            params = function['params']
            desc = function['desc']
            if name is None or name == "":
                continue
            if params is None or params == "":
                params = '{"type": "object", "properties": {}}'
            if desc is None:
                desc = ""
            functions.append({"name": name, "params": params, "desc": desc})
        if len(functions) > 0:
            assistant.tools['function'] = functions

    def clear(self, force=False):
        """
        Clears assistant data

        :param force: force clear data
        """
        id = self.window.config.data['assistant']

        if not force:
            self.window.ui.dialogs.confirm('assistant_clear', '', trans('confirm.assistant.clear'))
            return

        if id is not None and id != "":
            if self.assistants.has(id):
                assistant = self.assistants.get_by_id(id)
                assistant.reset()

        self.window.set_status(trans('status.assistant.cleared'))
        self.update()

    def select_file(self, idx):
        """
        Selects file

        :param idx: index of file
        """
        assistant_id = self.window.config.data['assistant']
        if assistant_id is None or assistant_id == "":
            return
        assistant = self.assistants.get_by_id(assistant_id)
        if assistant is None:
            return
        self.assistants.current_file = self.assistants.get_file_id_by_idx(assistant, idx)

    def sync_files(self, force=False):
        """
        Syncs files with API

        :param force: force sync files
        """
        if not force:
            self.window.ui.dialogs.confirm('assistant_import_files', '', trans('confirm.assistant.import_files'))
            return

        assistant_id = self.window.config.data['assistant']
        if assistant_id is None or assistant_id == "":
            return
        assistant = self.assistants.get_by_id(assistant_id)
        if assistant is None:
            return
        try:
            self.import_files(assistant)
        except Exception as e:
            print(e)
            self.window.ui.dialogs.alert(str(e))

    def clear_files(self, force=False):
        """
        Deletes all files

        :param force: force clear files
        """
        if not force:
            self.window.ui.dialogs.confirm('attachments_uploaded.clear', -1, trans('attachments_uploaded.clear.confirm'))
            return

        assistant_id = self.window.config.data['assistant']
        if assistant_id is None or assistant_id == "":
            return

        # delete all files
        if self.assistants.has(assistant_id):
            assistant = self.assistants.get_by_id(assistant_id)
            for file_id in list(assistant.files):
                try:
                    self.window.gpt.assistant_file_delete(assistant_id, file_id)
                    assistant.delete_file(file_id)
                except Exception as e:
                    self.window.ui.dialogs.alert(str(e))

                # delete file
                if assistant.has_file(file_id):
                    assistant.delete_file(file_id)
                # delete attachment
                if assistant.has_attachment(file_id):
                    assistant.delete_attachment(file_id)

            self.assistants.save()
            self.update()

    def delete(self, idx=None, force=False):
        """
        Deletes assistant

        :param idx: assistant index (row index)
        :param force: force delete without confirmation
        """
        if idx is not None:
            id = self.assistants.get_by_idx(idx)
            if id is not None and id != "":
                if self.assistants.has(id):
                    # if exists then show confirmation dialog
                    if not force:
                        self.window.ui.dialogs.confirm('assistant_delete', idx, trans('confirm.assistant.delete'))
                        return

                    # clear if this is current assistant
                    if id == self.window.config.data['assistant']:
                        self.window.config.data['assistant'] = None
                        self.window.config.data['assistant_thread'] = None

                    # delete in API
                    try:
                        self.window.gpt.assistant_delete(id)
                    except Exception as e:
                        self.window.ui.dialogs.alert(str(e))

                    self.assistants.delete(id)
                    self.update()
                    self.window.set_status(trans('status.assistant.deleted'))

    def config_toggle(self, id, value, section=None):
        """
        Toggles checkbox

        :param id: checkbox option id
        :param value: checkbox option value
        :param section: settings section
        """
        assistant_id = self.window.config.data['assistant']  # current assistant
        is_current = True
        if section == 'assistant.editor':
            assistant_id = self.window.config_option['assistant.id']  # editing assistant
            is_current = False
        self.update_field(id, value, assistant_id, is_current)

    def config_change(self, id, value, section=None):
        """
        Changes input value

        :param id: input option id
        :param value: input option value
        :param section: settings section
        """
        # validate filename
        if id == 'assistant.id':
            self.window.config_option[id].setText(value)

        assistant_id = self.window.config.data['assistant']  # current assistant
        is_current = True
        if section == 'assistant.editor':
            assistant_id = self.window.config_option['assistant.id']  # editing assistant
            is_current = False
        self.update_field(id, value, assistant_id, is_current)
        self.window.config_option[id].setText('{}'.format(value))

    def rename_file(self, idx):
        """
        Renames file

        :param mode: mode
        :param idx: selected attachment index
        """
        assistant_id = self.window.config.data['assistant']
        if assistant_id is None or assistant_id == "":
            return
        assistant = self.assistants.get_by_id(assistant_id)
        if assistant is None:
            return

        # get attachment ID by index
        file_id = self.assistants.get_file_id_by_idx(assistant, idx)

        # get attachment object by ID
        data = self.assistants.get_file_by_id(assistant, file_id)
        if data is None:
            return

        # set dialog and show
        self.window.dialog['rename'].id = 'attachment_uploaded'
        self.window.dialog['rename'].input.setText(data['name'])
        self.window.dialog['rename'].current = file_id
        self.window.dialog['rename'].show()
        self.update()

    def update_file_name(self, file_id, name):
        """
        Renames uploaded remote file name

        :param file_id: file_id
        :param name: new name
        """
        assistant_id = self.window.config.data['assistant']
        if assistant_id is None or assistant_id == "":
            self.close_rename_file()
            return
        assistant = self.assistants.get_by_id(assistant_id)
        if assistant is None:
            self.close_rename_file()
            return
        self.assistants.rename_file(assistant, file_id, name)
        self.close_rename_file()

    def close_rename_file(self):
        # close rename dialog and update attachments list
        self.window.dialog['rename'].close()
        self.update()

    def delete_file(self, idx, force=False):
        """
        Deletes file

        :param idx: file idx
        :param force: force delete without confirmation
        """
        if not force:
            self.window.ui.dialogs.confirm('attachments_uploaded.delete', idx, trans('attachments_uploaded.delete.confirm'))
            return

        # get current assistant
        assistant_id = self.window.config.data['assistant']
        if assistant_id is None or assistant_id == "":
            return
        assistant = self.assistants.get_by_id(assistant_id)
        if assistant is None:
            return

        # get attachment ID by index
        file_id = self.assistants.get_file_id_by_idx(assistant, idx)

        # delete file in API
        try:
            self.window.gpt.assistant_file_delete(assistant_id, file_id)
        except Exception as e:
            self.window.ui.dialogs.alert(str(e))
            return  # do not delete locally if not deleted in API

        # delete locally
        if self.assistants.has(assistant_id):
            need_save = False
            # delete file
            if assistant.has_file(file_id):
                assistant.delete_file(file_id)
                need_save = True
            # delete attachment
            if assistant.has_attachment(file_id):
                assistant.delete_attachment(file_id)
                need_save = True
            # save assistants and update assistants list
            if need_save:
                self.assistants.save()
                self.update()

    def clear_attachments(self, assistant):
        """
        Clears attachments

        :param assistant: assistant object
        """
        assistant.clear_attachments()
        self.assistants.save()
        self.update()

    def count_upload_attachments(self, attachments):
        """
        Counts uploaded attachments
        """
        num = 0
        for id in list(attachments):
            attachment = attachments[id]
            if not attachment.send:
                num += 1  # increment uploaded files counter
        return num

    def upload_attachments(self, mode, attachments):
        """
        Uploads attachments to assistant
        """
        # get current chosen assistant
        assistant_id = self.window.config.data['assistant']
        if assistant_id is None:
            return 0
        assistant = self.assistants.get_by_id(assistant_id)

        num = 0
        # loop on attachments
        for id in list(attachments):
            attachment = attachments[id]
            old_id = attachment.id  # tmp id

            # check if not already uploaded (ignore already uploaded files)
            if not attachment.send:
                print("Uploading file: {}".format(attachment.path))
                # check if file exists
                if not os.path.exists(attachment.path):
                    continue

                # upload local attachment file and get new ID (file_id)
                new_id = self.window.gpt.assistant_file_upload(assistant_id, attachment.path)
                if new_id is not None:
                    # mark as already uploaded
                    attachment.send = True
                    attachment.id = new_id
                    attachment.remote = new_id

                    # replace old ID with new one
                    self.window.controller.attachment.attachments.replace_id(mode, old_id, attachment)

                    # update assistant remote files list
                    assistant.files[new_id] = {
                        'id': new_id,
                        'name': attachment.name,
                        'path': attachment.path,
                    }

                    # update assistant attachments list
                    self.assistants.replace_attachment(assistant, attachment, old_id, new_id)
                    num += 1  # increment uploaded files counter

        # update assistants list
        self.assistants.save()

        # update attachments UI
        self.window.controller.attachment.update()

        # update uploaded list
        if num > 0:
            self.update_uploaded()  # update uploaded list UI

        return num

    def append_attachment(self, assistant, attachment):
        """
        Appends attachment to assistant

        :param attachment: attachment
        """
        # get current chosen assistant
        assistant.add_attachment(attachment)  # append attachment
        self.assistants.save()  # save assistants

    def goto_online(self):
        """Opens Assistants page"""
        webbrowser.open('https://platform.openai.com/assistants')

    def create_thread(self):
        """
        Creates assistant thread
        """
        thread_id = self.window.gpt.assistant_thread_create()
        self.window.config.data['assistant_thread'] = thread_id
        self.window.gpt.context.append_thread(thread_id)
        return thread_id

    def select_assistant_by_current(self):
        """Selects assistant by current"""
        assistant_id = self.window.config.data['assistant']
        items = self.window.controller.assistant.assistants.get_all()
        if assistant_id in items:
            idx = list(items.keys()).index(assistant_id)
            current = self.window.models['assistants'].index(idx, 0)
            self.window.data['assistants'].setCurrentIndex(current)

    def handle_message_files(self, msg):
        """
        Handles (download) message files

        :param msg: message
        :return:
        """
        num_downloaded = 0
        paths = []
        for file_id in msg.file_ids:
            path = self.window.controller.attachment.download(file_id)
            if path is not None:
                paths.append(path)
                num_downloaded += 1
        if num_downloaded > 0:
            # show alert with downloaded files
            msg = "Downloaded {} file(s): {}".format(num_downloaded, ", ".join(paths))
            self.window.ui.dialogs.alert(msg)

    def handle_run_messages(self, ctx):
        """
        Handles run messages

        :param ctx: context
        :return:
        """
        data = self.window.gpt.assistant_msg_list(ctx.thread)
        for msg in data:
            if msg.role == "assistant":
                ctx.set_output(msg.content[0].text.value)
                self.handle_message_files(msg)
                self.window.controller.input.handle_response(ctx, 'assistant', False)
                self.window.controller.input.handle_commands(ctx)
                break

    def handle_run(self, ctx):
        """
        Handles assistant's run

        :param ctx: context
        :return:
        """
        listener = AssistantRunThread(window=self.window, ctx=ctx)
        listener.updated.connect(self.handle_status)
        listener.destroyed.connect(self.handle_destroy)
        listener.started.connect(self.handle_started)

        self.thread_run = threading.Thread(target=listener.run)
        self.thread_run.start()
        self.thread_run_started = True

    @Slot(str, object)
    def handle_status(self, status, ctx):
        """
        Insert text to input and send

        :param status: status
        """
        print("Run status: {}".format(status))
        if status != "queued" and status != "in_progress":
            self.window.controller.input.unlock_input()  # unlock input
        if status == "completed":
            self.force_stop = False
            self.handle_run_messages(ctx)
            self.window.statusChanged.emit(trans('assistant.run.completed'))
        elif status == "failed":
            self.force_stop = False
            self.window.controller.input.unlock_input()
            self.window.statusChanged.emit(trans('assistant.run.failed'))

    @Slot()
    def handle_destroy(self):
        """
        Insert text to input and send
        """
        self.thread_run_started = False
        self.force_stop = False

    @Slot()
    def handle_started(self):
        """
        Handle listening started
        """
        print("Run: assistant is listening status...")
        self.window.statusChanged.emit(trans('assistant.run.listening'))


class AssistantRunThread(QObject):
    updated = Signal(object, object)
    destroyed = Signal()
    started = Signal()

    def __init__(self, window=None, ctx=None):
        """
        Run assistant run status check thread
        """
        super().__init__()
        self.window = window
        self.ctx = ctx
        self.check = True
        self.stop_reasons = ["cancelling", "cancelled", "failed", "completed", "expired", "requires_action"]

    def run(self):
        try:
            self.started.emit()
            while self.check \
                    and not self.window.is_closing \
                    and not self.window.controller.assistant.force_stop:
                status = self.window.gpt.assistant_run_status(self.ctx.thread, self.ctx.run_id)
                self.updated.emit(status, self.ctx)
                # finished or failed
                if status in self.stop_reasons:
                    self.check = False
                    self.destroyed.emit()
                    break
                time.sleep(1)
            self.destroyed.emit()
        except Exception as e:
            print(e)
            self.destroyed.emit()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.06 12:00:00                  #
# ================================================== #
import os

from langchain.llms import OpenAI
from langchain.chat_models import ChatOpenAI


class OpenAILLM:
    def __init__(self):
        self.id = "openai"

    def completion(self, config, options, stream=False):
        args = {}
        if 'args' in options:
            args = options['args']
        os.environ['OPENAI_API_KEY'] = config["api_key"]
        os.environ['OPENAI_API_TOKEN'] = config["api_key"]
        llm = OpenAI(**args)
        return llm

    def chat(self, config, options, stream=False):
        args = {}
        if 'args' in options:
            args = options['args']
        os.environ['OPENAI_API_KEY'] = config["api_key"]
        os.environ['OPENAI_API_TOKEN'] = config["api_key"]
        llm = ChatOpenAI(**args)
        return llm


# !/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.05 22:00:00                  #
# ================================================== #

from .model import Model
from .presets import Presets
from .plugins import Plugins
from .debug import Debug
from .settings import Settings
from .info import Info
from .input import Input
from .output import Output
from .context import Context
from .confirm import Confirm
from .ui import UI
from .launcher import Launcher
from .lang import Lang
from .image import Image
from .theme import Theme
from .audio import Audio
from .attachment import Attachment
from .notepad import Notepad
from .files import Files
from .assistant import Assistant
from .layout import Layout
from .camera import Camera


class Controller:
    def __init__(self, window=None):
        """
        Main controller

        :param window: main window object
        """
        self.window = window
        self.model = Model(window)
        self.presets = Presets(window)
        self.plugins = Plugins(window)
        self.debug = Debug(window)
        self.settings = Settings(window)
        self.info = Info(window)
        self.input = Input(window)
        self.output = Output(window)
        self.context = Context(window)
        self.confirm = Confirm(window)
        self.ui = UI(window)
        self.launcher = Launcher(window)
        self.lang = Lang(window)
        self.image = Image(window)
        self.theme = Theme(window)
        self.audio = Audio(window)
        self.attachment = Attachment(window)
        self.notepad = Notepad(window)
        self.files = Files(window)
        self.assistant = Assistant(window)
        self.layout = Layout(window)
        self.camera = Camera(window)

    def setup(self):
        """Setups controller"""
        # init material theme
        self.layout.setup()

        # setup all controllers
        self.lang.setup()
        self.model.setup()
        self.assistant.setup()
        self.input.setup()
        self.output.setup()
        self.context.setup()
        self.ui.setup()
        self.info.setup()
        self.launcher.setup()
        self.audio.setup()
        self.attachment.setup()
        self.notepad.setup()
        self.camera.setup_settings()
        self.image.setup()

    def setup_plugins(self):
        """Setup plugins"""
        # setup plugins settings
        self.plugins.setup_settings()

    def update(self):
        """On update"""
        self.camera.update()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.13 12:00:00                  #
# ================================================== #
import datetime
import os
import threading
import cv2
from PySide6.QtCore import Slot

from PySide6.QtGui import QImage, QPixmap, Qt
from ..camera import CameraThread
from ..utils import trans


class Camera:
    def __init__(self, window=None):
        """
        Camera controller

        :param window: main window
        """
        self.window = window
        self.thread = None
        self.frame = None
        self.thread_started = False
        self.is_capture = False
        self.stop = False
        self.auto = False

    def setup(self):
        """
        Setup camera
        """
        if self.is_capture and not self.thread_started:
            self.start()

    def start(self):
        """
        Start camera thread
        """
        if self.thread_started:
            return

        # prepare thread
        self.stop = False
        thread = CameraThread(window=self.window)
        thread.finished.connect(self.handle_stop)
        thread.stopped.connect(self.handle_stop)

        # run thread
        self.thread = threading.Thread(target=thread.run)
        self.thread.start()
        self.thread_started = True

    def stop_capture(self):
        """
        Stop camera capture thread
        """
        if not self.thread_started:
            return

        self.stop = True

    @Slot()
    def handle_stop(self):
        """
        On capture stopped signal
        """
        self.thread_started = False
        self.thread = None
        self.hide_camera(False)

    def blank_screen(self):
        """
        Make and set blank screen
        """
        self.window.data['video.preview'].video.setPixmap(QPixmap.fromImage(QImage()))

    def update(self):
        """
        Update camera frame
        """
        if self.thread is None \
                or not self.thread_started \
                or self.frame is None \
                or not self.is_capture:
            return

        # scale and update frame
        width = self.window.data['video.preview'].video.width()
        image = QImage(self.frame, self.frame.shape[1], self.frame.shape[0],
                       self.frame.strides[0], QImage.Format_RGB888)
        pixmap = QPixmap.fromImage(image)
        scaled_pixmap = pixmap.scaled(width, pixmap.height(),
                                      Qt.KeepAspectRatio, Qt.SmoothTransformation)
        self.window.data['video.preview'].video.setPixmap(scaled_pixmap)

    def manual_capture(self):
        """
        Capture frame via click on video output
        """
        if not self.is_auto():
            if not self.capture_frame(True):
                self.window.statusChanged.emit(trans("vision.capture.manual.captured.error"))
        else:
            self.window.statusChanged.emit(trans('vision.capture.auto.click'))

    def capture_frame(self, switch=True):
        """
        Capture frame and save it as attachment

        :param switch: True if switch to attachments tab (tmp: disabled)
        """

        # clear attachments before capture if needed
        if self.window.controller.attachment.is_capture_clear():
            self.window.controller.attachment.clear(True)

        # capture frame
        try:
            # prepare filename
            now = datetime.datetime.now()
            dt = now.strftime("%Y-%m-%d_%H-%M-%S")
            name = 'cap-' + dt
            path = os.path.join(self.window.config.path, 'capture', name + '.jpg')

            # capture frame
            compression_params = [cv2.IMWRITE_JPEG_QUALITY, self.window.config.data['vision.capture.quality']]
            frame = cv2.cvtColor(self.frame, cv2.COLOR_BGR2RGB)
            cv2.imwrite(path, frame, compression_params)
            mode = self.window.config.data['mode']

            # make attachment
            dt_info = now.strftime("%Y-%m-%d %H:%M:%S")
            title = trans('vision.capture.name.prefix') + ' ' + name
            title = title.replace('cap-', '').replace('_', ' ')
            self.window.controller.attachment.attachments.new(mode, title, path, False)
            self.window.controller.attachment.attachments.save()
            self.window.controller.attachment.update()

            # show last capture time in status
            self.window.statusChanged.emit(trans("vision.capture.manual.captured.success") + ' ' + dt_info)

            # switch to attachments tab if needed (tmp: disabled)
            if switch:
                pass
                # self.window.tabs['input'].setCurrentIndex(1)  # 1 = index of attachments tab
            return True
        except Exception as e:
            print("Frame capture exception", e)
            self.window.statusChanged.emit(trans('vision.capture.error'))
        return False

    def show_camera(self):
        """
        Show camera
        """
        if self.is_capture:
            self.window.data['video.preview'].setVisible(True)

    def hide_camera(self, stop=True):
        """
        Hide camera

        :param stop: True if stop capture thread
        """
        self.window.data['video.preview'].setVisible(False)

        if stop:
            self.stop_capture()

    def enable_capture(self):
        """
        Enable capture
        """
        if self.window.config.data['mode'] != 'vision':
            return

        self.is_capture = True
        self.window.config.data['vision.capture.enabled'] = True
        self.window.data['video.preview'].setVisible(True)
        if not self.thread_started:
            self.start()

    def disable_capture(self):
        """
        Disable capture
        """
        if self.window.config.data['mode'] != 'vision':
            return

        self.is_capture = False
        self.window.config.data['vision.capture.enabled'] = False
        self.window.data['video.preview'].setVisible(False)
        self.stop_capture()
        self.blank_screen()

    def toggle(self, state):
        """
        Toggle camera

        :param state: state
        """
        if state:
            self.enable_capture()
        else:
            self.disable_capture()

        self.window.statusChanged.emit('')

    def enable_auto(self):
        """
        Enable capture
        """
        if self.window.config.data['mode'] != 'vision':
            return

        self.auto = True
        self.window.config.data['vision.capture.auto'] = True
        self.window.data['video.preview'].label.setText(trans("vision.capture.auto.label"))

    def disable_auto(self):
        """
        Disable capture
        """
        if self.window.config.data['mode'] != 'vision':
            return

        self.auto = False
        self.window.config.data['vision.capture.auto'] = False
        self.window.data['video.preview'].label.setText(trans("vision.capture.label"))

    def toggle_auto(self, state):
        """
        Toggle camera

        :param state: state (True/False)
        """
        if state:
            self.enable_auto()
        else:
            self.disable_auto()

        self.window.statusChanged.emit('')

    def is_enabled(self):
        """
        Check if camera is enabled

        :return: True if enabled, False otherwise
        """
        return self.is_capture

    def is_auto(self):
        """
        Check if camera is enabled

        :return: True if enabled, False otherwise
        """
        return self.auto

    def setup_settings(self):
        """
        Update layout checkboxes
        """
        if self.window.config.data['vision.capture.enabled']:
            self.is_capture = True
            self.window.data['vision.capture.enable'].setChecked(True)
        else:
            self.is_capture = False
            self.window.data['vision.capture.enable'].setChecked(False)

        if self.window.config.data['vision.capture.auto']:
            self.auto = True
            self.window.data['vision.capture.auto'].setChecked(True)
        else:
            self.auto = False
            self.window.data['vision.capture.auto'].setChecked(False)

        # update label
        if not self.window.config.data['vision.capture.auto']:
            self.window.data['video.preview'].label.setText(trans("vision.capture.label"))
        else:
            self.window.data['video.preview'].label.setText(trans("vision.capture.auto.label"))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.14 15:00:00                  #
# ================================================== #

class BasePlugin:
    def __init__(self):
        self.id = ""
        self.name = ""
        self.description = ""
        self.options = {}
        self.initial_options = {}
        self.window = None
        self.enabled = False
        self.order = 0

    def setup(self):
        """
        Returns available config options

        :return: config options
        """
        return self.options

    def add_option(self, name, type, value=None, label="", description="",
                   tooltip=None, min=None, max=None, multiplier=1, step=1, slider=False,
                   keys=None, advanced=False, secret=False, persist=False):
        """
        Adds option

        :param name: Option name (ID, key)
        :param type: Option type (text, textarea, bool, int, float, dict)
        :param value: Option value
        :param label: Option label
        :param description: Option description
        :param tooltip: Option tooltip
        :param min: Option min value
        :param max: Option max value
        :param multiplier: Option float value multiplier
        :param step: Option step value (for slider)
        :param slider: Option slider (True/False)
        :param keys: Option keys (for dict type)
        :param advanced: Option advanced (True/False)
        :param secret: Option secret (True/False)
        :param persist: Option persist (True/False)
        """
        if tooltip is None:
            tooltip = description

        option = {
            "type": type,
            "value": value,
            "label": label,
            "description": description,
            "tooltip": tooltip,
            "min": min,
            "max": max,
            "multiplier": multiplier,
            "step": step,
            "slider": slider,
            "keys": keys,
            "advanced": advanced,
            "secret": secret,
            "persist": persist,
        }
        self.options[name] = option

    def has_option(self, name):
        """
        Checks if option exists

        :param name: Option name
        :return: True if exists
        """
        return name in self.options

    def get_option(self, name):
        """
        Returns option

        :param name: Option name
        :return: Option
        """
        if self.has_option(name):
            return self.options[name]

    def attach(self, window):
        """
        Attaches window

        :param window: Window
        """
        self.window = window

    def on_user_send(self, text):
        """
        Event: On user send text

        :param text: Text
        :return: Text
        """
        return text

    def on_ctx_begin(self, ctx):
        """
        Event: On new context begin

        :param ctx: Context
        :return: Context
        """
        return ctx

    def on_ctx_end(self, ctx):
        """
        Event: On context end

        :param ctx: Context
        :return: Context
        """
        return ctx

    def on_system_prompt(self, prompt):
        """
        Event: On prepare system prompt

        :param prompt: Prompt
        :return: Prompt
        """
        return prompt

    def on_ai_name(self, name):
        """
        Event: On set AI name

        :param name: Name
        :return: Name
        """
        return name

    def on_user_name(self, name):
        """
        Event: On set user name

        :param name: Name
        :return: Name
        """
        return name

    def on_enable(self):
        """Event: On plugin enable"""
        pass

    def on_disable(self):
        """Event: On plugin disable"""
        pass

    def on_input_before(self, text):
        """
        Event: Before input

        :param text: Text
        """
        return text

    def on_ctx_before(self, ctx):
        """
        Event: Before ctx

        :param ctx: Text
        """
        return ctx

    def on_ctx_after(self, ctx):
        """
        Event: After ctx

        :param ctx: ctx
        """
        return ctx
